!> evaluation of stiff matrix and the appropriate submatrix
!> solution of algebraic problem by an iterative method
module o_matrix
  use o_mesh


  implicit none

contains

  ! evaluation of the stiff matrix for interior and boundary nodes
  subroutine STIFF_MATRIX(grid, mS  )
    type(mesh), target, intent(inout) :: grid
    real, dimension(1:grid%npoin, 1:grid%npoin), intent(out) :: mS  ! = stiff matrix
    real, dimension(:,:), pointer :: xp
    integer, dimension(:, : ), pointer :: nodes
    real, dimension(:,:), allocatable :: Dphi, xi
    real :: D  ! local variables
    integer :: i,j,k,ki, kj, nelem, npoin

    nelem = grid%nelem
    npoin = grid%npoin

    xp => grid%xp(1:npoin, 1:2)
    nodes => grid%nodes(1:nelem,1:3)

    allocate(Dphi(1:3, 1:2), xi(1:3, 1:2) )
    mS(:, :) = 0.

    do k=1,nelem
       ! coordinates of the element
       xi(1:3, 1:2) =  xp(nodes(k, 1:3), 1:2 )

       ! double area of element
       D =    xi(1,1) * (xi(2,2)-xi(3,2))  &
            + xi(2,1) * (xi(3,2)-xi(1,2)) &
            + xi(3,1) * (xi(1,2)-xi(2,2))

       grid%area(k) = D/2

       call GRADIENTS( xi(1:3,1), xi(1:3, 2), Dphi(1:3, 1:2) )

       do ki=1,3
          i = nodes(k,ki) 
          do kj=1,3
             j = nodes(k,kj)
             if(D.le. 0) then
                print*,'?? zero area = ', D
                print*,xi(1,1:2)
                print*,xi(2,1:2)
                print*,xi(3,1:2)
                stop 
             endif

             mS(i,j) = mS(i,j) + (Dphi(ki,1)* Dphi(kj,1) +  Dphi(ki,2)* Dphi(kj,2) )/D / 2
          enddo
       enddo
    enddo

  end subroutine STIFF_MATRIX

  ! evaluation of the stiff matrix for interior and boundary nodes
  subroutine STIFF_MATRIX_SPARSE(grid )
    type(mesh), target, intent(inout) :: grid
    real, dimension(:,:), pointer :: xp
    integer, dimension(:, : ), pointer :: nodes
    real, dimension(:,:), allocatable :: Dphi, xi
    real :: D  ! local variables
    integer :: i,j,k,ki, kj, nelem, npoin, kk, nzero

    grid%sparse(:) = 0.
    
    nelem = grid%nelem
    npoin = grid%npoin
    nzero = grid%nzero

    xp => grid%xp(1:npoin, 1:2)
    nodes => grid%nodes(1:nelem,1:3)

    allocate(Dphi(1:3, 1:2), xi(1:3, 1:2) )
    !mS(:, :) = 0.

    do k=1,nelem
       ! coordinates of the element
       xi(1:3, 1:2) =  xp(nodes(k, 1:3), 1:2 )

       ! double area of element
       D =    xi(1,1) * (xi(2,2)-xi(3,2))  &
            + xi(2,1) * (xi(3,2)-xi(1,2)) &
            + xi(3,1) * (xi(1,2)-xi(2,2))

       grid%area(k) = D/2

       call GRADIENTS( xi(1:3,1), xi(1:3, 2), Dphi(1:3, 1:2) )

       do ki=1,3
          i = nodes(k,ki) 
          do kj=1,3
             j = nodes(k,kj)
             if(D.le. 0) then
                print*,'?? zero area = ', D
                print*,xi(1,1:2)
                print*,xi(2,1:2)
                print*,xi(3,1:2)
                stop 
             endif

             kk = Find_Sparce_index(npoin, nzero, grid%irow, grid%icol, i,j)
             grid%sparse(kk) = grid%sparse(kk) &
                  + (Dphi(ki,1)* Dphi(kj,1) +  Dphi(ki,2)* Dphi(kj,2) )/D / 2
             
             !print*,'###', i,j,kk
             !mS(i,j) = mS(i,j) + (Dphi(ki,1)* Dphi(kj,1) +  Dphi(ki,2)* Dphi(kj,2) )/D / 2
          enddo
       enddo
    enddo

  end subroutine STIFF_MATRIX_SPARSE


  subroutine GRADIENTS(x, y, Dphi )
    real, intent(in) :: x(1:3), y(1:3)
    real, intent(out) :: Dphi(1:3, 1:2)  
    ! Dphi(i,j) = derivative of i-th function with respect x_j (multiplied by 2|K|)

    Dphi(1,1) = y(2) - y(3)
    Dphi(1,2) = x(3) - x(2)

    Dphi(2,1) = y(3) - y(1)
    Dphi(2,2) = x(1) - x(3)

    Dphi(3,1) = y(1) - y(2)
    Dphi(3,2) = x(2) - x(1)
  end subroutine GRADIENTS


  ! assembling of the smaller matrix, including only interior node
   subroutine Assemble_small_matrix(grid, mS, mA ) 
    type(mesh), target, intent(inout) :: grid
    integer, dimension(:,:), pointer :: ip  ! interior / boundary
    real, dimension(1:grid%npoin, 1:grid%npoin), intent(in) :: mS  ! = global stiff matrix
    real, dimension(1:grid%ipoin, 1:grid%ipoin), intent(inout) :: mA  ! = actual stiff matrix
    integer :: i,j,k, l, npoin

    npoin = grid%npoin

    ip => grid%ip(1:npoin, 1:2)

    mA = 0.

    k = 0 ! local row index
    do i=1, npoin
       if(ip(i, 1) >= 0) then  
          k = k + 1

          l = 0 ! local column index

          do j=1,npoin
             if(ip(j, 1) >= 0) then
                l = l+1
                mA(k,l) = mS(i, j)
             endif
          enddo

       endif
    enddo


  end subroutine Assemble_small_matrix


  !> solution of linear algebraic problem by Jacobi method
  subroutine SolveMatrixProblem(N, mA, f, v )
    integer, intent(in) :: N ! number of equations
    real, dimension(1:N, 1:N), intent(in) :: mA !  matrix
    real, dimension(1:N), intent(in) :: f     ! RHS 
    real, dimension(1:N), intent(inout) :: v  ! solution 
    real, dimension(:),allocatable :: v1
    integer :: i,j, max_iter
    real :: resid
    real :: tol = 1E-12

    allocate(v1(1:N) )

    max_iter = 10 * N

    print*
    print*,'  Problems size = ', N
    do i=1, max_iter
       v1(1:N) = f(1:N) - matmul(mA(1:N, 1:N), v(1:N) )

       resid = norm2(v1(1:N))

       if(i==1 .or. mod(i, 100) == 0 .or. resid <= tol) &
            write(*,'(a5,i7, a10, es12.4)') 'iters=', i, ', residuum = ',resid

       if(resid <= tol) goto 10

       do j=1,N
          v(j) = v(j) + v1(j) / mA(j,j)
       enddo

    end do

10  continue
    print*,'algebraic residuum = ', resid

    deallocate(v1)

  end subroutine SolveMatrixProblem


    !> solution of linear algebraic problem by Jacobi method
  subroutine SolveMatrixProblem_Sparse(grid, N, f, v )
    type(mesh), target, intent(inout) :: grid
    integer, intent(in) :: N ! number of equations
    !!real, dimension(1:N, 1:N), intent(in) :: mA !  matrix  ! NOT USED
    real, dimension(1:N), intent(in) :: f     ! RHS 
    real, dimension(1:N), intent(inout) :: v  ! solution 
    real, dimension(:),allocatable :: v1, v2
    integer, dimension(:,:), pointer :: ip  ! interior / boundary
    integer :: i,j, max_iter, npoin, is
    real :: resid
    real :: tol = 1E-12

    npoin = grid%npoin
    ip => grid%ip(1:npoin, 1:2)

    allocate(v1(1:N), v2(1:N) )

    max_iter = 10 * N

    print*
    print*,'  Problems size = ', N
    do i=1, max_iter
       call product_Ax(grid, N, v(1:N), v2(1:N) )

       !write(*,'(a4,200es12.4)') 'new:', v2(1:N)
       !write(*,'(a4,200es12.4)') 'old:', matmul(mA(1:N, 1:N), v(1:N) )
       !print*,'________________________'
       !stop "30dh3ihd3"
       
       v1(1:N) = f(1:N) - v2(1:N)
       !!v1(1:N) = f(1:N) - matmul(mA(1:N, 1:N), v(1:N) )

       resid = norm2(v1(1:N))

       if(i==1 .or. mod(i, 100) == 0 .or. resid <= tol) &
            write(*,'(a5,i7, a10, es12.4)') 'iters=', i, ', residuum = ',resid

       if(resid <= tol) goto 10

       ! original formula
       !do j=1,N
       !   v(j) = v(j) + v1(j) / mA(j,j)
       !enddo

       ! sparce variant 
       do is = 1, npoin
          if(ip(is, 1) >= 0) then
             j = ip(is, 2)
             v(j) = v(j) + v1(j) / grid%sparse(grid%irow(is))  ! first row entry is the diagonal one
          endif
       enddo
       

    end do

10  continue
    print*,'algebraic residuum = ', resid

    deallocate(v1, v2)

  end subroutine SolveMatrixProblem_Sparse


  subroutine product_Ax(grid, N, v1, v2)
    type(mesh), target, intent(inout) :: grid
    integer, intent(in) :: N ! number of equations
    real, dimension(1:N), intent(in) :: v1     ! input
    real, dimension(1:N), intent(inout) :: v2  ! output
    integer :: i, is, j, js, ks, ipoin, npoin
    integer, dimension(:,:), pointer :: ip  ! interior / boundary

    npoin = grid%npoin
    ip => grid%ip(1:npoin, 1:2)

    v2(:) = 0.
    
    do is = 1, npoin
       if(ip(is, 1) >=0 ) then
          i = ip(is, 2)

          do ks = grid%irow(is), grid%irow(is+1) -1

             js = grid%icol(ks)

             if(ip(js, 1) >=0) then
                j = ip(js, 2)

                !print*,'????', i, ks, js, j
                v2(i) = v2(i) + grid%sparse(ks) * v1(j)
                
             endif
          enddo
       endif
    enddo
    
    
  end subroutine product_Ax

  
end module o_matrix


