!> triangular mesh 
module o_mesh

  implicit none


  !> boundary edge, from input and the curved representation, if any
  type, public :: mesh
     integer :: nelem    ! number of elements
     integer :: npoin    ! number of nodes
     integer :: ipoin    ! number of interior nodes
     integer :: nbelm    ! number of boundary elements
     integer :: nbc      ! number of boundary components
     integer:: maxdeg    ! maximal number of edges for one node
     integer:: nzero     ! number of non-zero elements
     
     real, allocatable, dimension(:,:) :: xp   ! coordinates of nodes
     integer, allocatable, dimension(:,:) :: nodes ! indexes of vertices of elements
     integer, allocatable, dimension(:,:) :: lbn   ! indexes of boundary edges
     integer, allocatable, dimension(:,:) :: ip ! interior or boundary nodes?
     real, allocatable, dimension(:) :: area  ! areas of elements
     real, allocatable, dimension(:) :: b_edge  ! only for Neumann nodes

     integer, allocatable, dimension(:,:) :: icyc
     integer, allocatable, dimension(:) :: irow
     integer, allocatable, dimension(:) :: icol
     real, allocatable, dimension(:) :: sparse
  end type mesh


  public :: Read_mesh


contains

  ! readin of the mesh from the file and setting of arrays
  subroutine Read_mesh(trifile, grid, sparse)
    character(len=*), intent( in) :: trifile
    type(mesh), target, intent(inout) :: grid
    character(len=1), intent(in):: sparse

    real :: r1, r2, r3, r4
    integer, dimension(:,:), pointer :: ip
    integer :: i, i1,i2,i3,i4, itri
    integer :: npoin, nelem, nbelm, nbc, ipoin
    real :: val


    ! reading of mesh
    itri = 10
    open(itri, file=trifile, status='OLD')

    read(itri,*)  npoin, nelem, nbelm, grid%nbc
    grid%npoin = npoin
    grid%nelem = nelem
    grid%nbelm = nbelm

    allocate(grid%xp(1:npoin, 1:2) )
    allocate(grid%nodes(1:nelem, 1:3) )
    allocate(grid%lbn(1:nbelm, 1:3) )
    allocate(grid%ip(1:npoin, 1:2) )

    allocate(grid%area(1:nelem), source = 0.0 ) ! computed in subroutine STIFF_MATRIX


    ! skipping of data
    read(itri, *) r1, r2, i1, i2, r3, r4, i3, i4

    ! coordinates of vertices
    do i=1,npoin
       read(itri, *) grid%xp(i, 1:2)
    enddo

    ! indexes of vertices of elements 
    do i=1,nelem
       read(itri, *) grid%nodes(i, 1:3)
       !print*,'nodes:',i, nodes(i, 1:3)
    enddo

    ! indexes of vertices of boundary edges + index of component
    do i=1,nbelm
       read(itri, *) grid%lbn(i, 1:3)
    enddo


    ip => grid%ip(1:npoin, 1:2)

    ! setting of array ip: ip(i) == 1 the -th node is interior
    ! ip(i, 1) == 1  => interior node
    ! ip(i, 1) == 0  => node on the Neumann part of boundary 
    ! ip(i, 1) == -1  => node on the Dirichlet part of boundary

    ip(:,1) = 1
    ip(:,2) = 0
    ! first all boundary nodes
    do i=1,nbelm
       ! i1 and i2 ... verices of an boundary edge
       i1 = grid%lbn(i,1)
       i2 = grid%lbn(i,2)

       ip(i1, 1) = 0  ! Boundary node
       ip(i2, 1) = 0
    enddo

    ! only Dirichlet nodes
    do i=1,nbelm
       ! i1 and i2 ... verices of an boundary edge
       i1 = grid%lbn(i,1)
       i2 = grid%lbn(i,2)
       i3 = grid%lbn(i,3)  ! number of boundary component

       ! if i3 == 2 then Neumann BC, otherwise Dirichlet
       if( i3 /= 2) then
       !if( i3 /= -2) then   ! NO NEUMANN BC
          ip(i1, 1) =  - 1  ! Boundary node
          ip(i2, 1) = - 1
       endif
    enddo

    ! setting the number of interior and Neumann nodes = number of unknowns
    ipoin = 0
    do i=1,npoin

       if(ip(i, 1) >= 0) then   ! Neumann or interior
          ipoin = ipoin + 1
          ip(i, 2) = ipoin
       endif

    enddo

    grid%ipoin = ipoin
    print*,"Number of interior or Neumann nodes = ", ipoin

    close(itri)

    ! only for Neumann nodes mean value of lengths of adjacent edges
    allocate(grid%b_edge(1:npoin), source = 0.0 )
    do i=1,nbelm
       ! i1 and i2 ... verices of an boundary edge
       i1 = grid%lbn(i,1)
       i2 = grid%lbn(i,2)

       ! 1/2 of the length of the edge
       val = norm2( grid%xp(i1, 1:2)  - grid%xp(i2, 1:2) ) / 2

       if(ip(i1, 1) == 0)  grid%b_edge(i1) = grid%b_edge(i1) + val
       if(ip(i2, 1) == 0)  grid%b_edge(i2) = grid%b_edge(i2) + val
    enddo


    if(sparse == 'S') call SetSparcity(grid)
    
  end subroutine Read_mesh

  subroutine SetSparcity( grid)
    type(mesh), target, intent(inout) :: grid
    integer, allocatable, dimension(:,:) :: icyc
    integer, dimension(:,:), pointer :: ip

    integer :: i,l, j, j1, m,k, k1, maxdeg, npoin, nelem, nzero

    
    maxdeg = 15
    npoin = grid%npoin
    nelem = grid%nelem

    allocate(icyc(1:npoin, 0:maxdeg), source = 0 )

    ! initialization
    do i=1,npoin
       icyc(i,0) = 1
       icyc(i,1) = i
    enddo

    do i=1,nelem
       do l=1, 3
          k = grid%nodes(i, l)
          do j=1, 3
             if(j /= l) then
                k1 = grid%nodes(i, j)
                do m=1, icyc(k, 0)
                   if(icyc(k,m) == k1) goto 100
                enddo
                ! adding a new node
                icyc(k, 0) = icyc(k,0) + 1

                if(icyc(k, 0) > maxdeg)   stop "out of range!!!"
                   
                icyc(k, icyc(k,0)) = k1

100             continue

             endif
          enddo
       enddo
    enddo

    
    ! 
    grid%nzero = sum(icyc(:,0) )
    grid%maxdeg = maxval(icyc(:,0) )
    
    print*,'Maxval = ',grid%maxdeg, ',    nonzeros = ', grid%nzero

    maxdeg = grid%maxdeg
    allocate(grid%icyc(1:npoin, 0:maxdeg), source = 0 )
    
    grid%icyc(1:npoin, 0:maxdeg) = icyc(1:npoin, 0:maxdeg)

    deallocate(icyc)
    
!WR    do i=1, npoin
!WR       write(*,'(i5,a2,i5,a2,30i5)') i,'|',grid%icyc(i,0),':',grid%icyc(i,1:grid%icyc(i,0))
!WR    enddo

    ! setting of the sparce format
    !ip => grid%ip(1:npoin, 1:2)

    allocate(grid%irow(1:grid%npoin+1) )
    allocate(grid%icol(1:grid%nzero) )
    allocate(grid%sparse(1:grid%nzero) )

    grid%irow(1) = 1

    nzero = 0
    do i = 1,npoin
       do j=1,  grid%icyc(i, 0)
          nzero = nzero + 1
          grid%icol(nzero) = grid%icyc(i, j)
       enddo
       grid%irow(i+1) = nzero + 1

    enddo

!WR    write(*,*) 
!WR    write(*,'(a8,300i5)') 'irow:', grid%irow(:)
!WR   write(*,'(a8,300i5)') 'icol:', grid%icol(:)
!WR    write(*,*)

    !stop

  end subroutine SetSparcity

  function Find_Sparce_index(npoin, nzero, irow, icol, i,j) result (k)
    integer :: k
    integer, intent(in) :: npoin, nzero, i, j
    integer, dimension(1:npoin+1), intent(in) :: irow
    integer, dimension(1:nzero), intent(in) :: icol
    integer :: m,n

    
    k = 0
    do m=irow(i), irow(i+1) -1
       if(icol(m) == j) then
          k = m
          return
       endif
    enddo
    if(k==0) then
       print*,'index does not found:' , i, j, icol(i), icol(i+1) -1
       stop
    endif

  end function Find_Sparce_index
end module o_mesh



